/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, FRAUNHOFER HHI
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of FRAUNHOFER HHI
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TDecBinCoderPIPE.cpp
    \brief    binary entropy decoder for PIPE
*/

#include "TDecBinCoderPIPE.h"

#ifdef GET_TIME
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

extern double DecEntTime;
extern double DecBitTime;
extern double DecEPTime;

#endif //GET_TIME

#if PIPE_LOW_DELAY_OPTION
V2VDecoder::V2VDecoder( const UInt uiPIPEId, UInt* puiMaxDelayBitCounter )
#else
V2VDecoder::V2VDecoder( const UInt uiPIPEId )
#endif
: m_uiPIPEId              ( uiPIPEId )
, m_paucStateTransition   ( TDecPIPETables::sm_StateTransition[ m_uiPIPEId ] )
, m_pauiCodeword          ( TDecPIPETables::sm_Codeword       [ m_uiPIPEId ] )
, m_pcTComBitstream       ( 0 )
#if PIPE_LOW_DELAY_OPTION
, m_uiMaxCodewordLength   ( TDecPIPETables::sm_MaxTableSize   [ m_uiPIPEId ] )
, m_uiMaxDelayInBits      ( UInt(-1) >> 1 )
, m_puiMaxDelayBitCounter ( puiMaxDelayBitCounter )
#endif
#if PIPE_V2V_V2F
, m_uiReadLength          ( TDecPIPETables::sm_CWFixLength  [ m_uiPIPEId ] )
#endif //PIPE_V2V_V2F
{
}

#if PIPE_LOW_DELAY_OPTION
Void
V2VDecoder::initDelay( UInt uiMaxDelayInBits )
{
  m_uiMaxDelayInBits = ( uiMaxDelayInBits ? uiMaxDelayInBits : ( UInt(-1) >> 1 ) );
}
#endif

Void
V2VDecoder::start()
{
  m_uiValue               = 0;
  m_uiLength              = 0;
#if PIPE_V2V_V2F
  m_uiLength_Trm = 0;
#endif //PIPE_V2V_V2F
#if PIPE_LOW_DELAY_OPTION
  m_uiMaxBitCounterValue  = m_uiMaxDelayInBits;
#endif
}


#if    PIPE_V2V_V2F
UInt
V2VDecoder::getNextBin()
{
#if PIPE_LOW_DELAY_OPTION
  if( m_uiLength == 0 || *m_puiMaxDelayBitCounter > m_uiMaxBitCounterValue )
  {
    m_uiMaxBitCounterValue    = *m_puiMaxDelayBitCounter + m_uiMaxDelayInBits;
    *m_puiMaxDelayBitCounter += m_uiMaxCodewordLength;
    switch ( m_uiPIPEId ){
      case PIPE_EP_CODER:
       xReadNextCodeword_EP();  break;
      case PIPE_TRM_CODER:
       xReadNextCodeword_Trm(); break;
      default:
       xReadNextCodeword();     break;
    }
  }
#else
  if( m_uiLength == 0 )
  {
    switch ( m_uiPIPEId ){
      case PIPE_EP_CODER:
       xReadNextCodeword_EP();  break;
      case PIPE_TRM_CODER:
       xReadNextCodeword_Trm(); break;
      default:
       xReadNextCodeword();     break;
    }
  }
#endif
  m_uiLength--;
  return ( ( m_uiValue >> m_uiLength ) & 1 );
}


#ifdef PIPE_V2F_BYPASS_MULTIREAD
__inline UInt
V2VDecoder::getNextBin_EP_multi( UInt uiLength )
{
#if PIPE_LOW_DELAY_OPTION
  if( m_uiLength < uiLength || *m_puiMaxDelayBitCounter > m_uiMaxBitCounterValue )
  {
    m_uiMaxBitCounterValue    = *m_puiMaxDelayBitCounter + m_uiMaxDelayInBits;
    *m_puiMaxDelayBitCounter += m_uiMaxCodewordLength;
    xReadNextCodeword_EP();
  }
#else
  while( m_uiLength < uiLength )
  {
    UInt    uiBit;
    m_pcTComBitstream->readV2F( uiBit );
    m_uiValue = (m_uiValue << 8) + uiBit;
    m_uiLength += 8;
  }
#endif
  m_uiLength -= uiLength;
  return ( ( m_uiValue >> m_uiLength ) & (0xffffff >> (24 - uiLength)) );
}
#endif //PIPE_V2F_BYPASS_MULTIREAD

#else //PIPE_V2V_V2F


UInt
V2VDecoder::getNextBin()
{
#if PIPE_LOW_DELAY_OPTION
  if( m_uiLength == 0 || *m_puiMaxDelayBitCounter > m_uiMaxBitCounterValue )
  {
    m_uiMaxBitCounterValue    = *m_puiMaxDelayBitCounter + m_uiMaxDelayInBits;
    *m_puiMaxDelayBitCounter += m_uiMaxCodewordLength;
    xReadNextCodeword();
  }
#else
  if( m_uiLength == 0 )
  {
    xReadNextCodeword();
  }
#endif
  m_uiLength--;
#ifdef PIPE_V2F_8BITS
  if( m_uiPIPEId == PIPE_TRM_CODER ){
    return m_uiLength ? 0 : (UInt)m_uiValue;
  }
  else
    return ( ( m_uiValue >> m_uiLength ) & 1 );
#else //PIPE_V2F_8BITS
  return ( ( m_uiValue >> m_uiLength ) & 1 );
#endif //PIPE_V2F_8BITS
}
#endif //PIPE_V2V_V2F


#ifdef WATCH_PIPE_2
UInt CWID = 0;
#endif //WATCH_PIPE_2

#if    PIPE_V2V_V2F
__inline Void
V2VDecoder::xReadNextCodeword_EP()
{
  UInt    uiBit;
  m_pcTComBitstream->readV2F( uiBit );
  m_uiValue = uiBit;
  m_uiLength = 8;
}


__inline Void
V2VDecoder::xReadNextCodeword_Trm()
{
  if( m_uiLength_Trm == 0 ){
    m_pcTComBitstream->readV2F( m_uiLength_Trm );
   if(m_uiLength_Trm == 255){
      m_uiValue_Trm = 0;
    }
    else{
      m_uiLength_Trm++;
      m_uiValue_Trm = 1;
    }
  }
  if( m_uiLength_Trm <= 64 )
  {
    m_uiLength = m_uiLength_Trm;
    m_uiValue  = m_uiValue_Trm;
    m_uiLength_Trm = 0;
  }
  else
  {
    m_uiLength     = 64;
    m_uiValue      = 0;
    m_uiLength_Trm -= 64;
  }
}

__inline Void
V2VDecoder::xReadNextCodeword()
{
  UInt    uiBit;
  UInt64  ui64CW_Top, ui64CW_Bottom;
  m_pcTComBitstream->readV2F( uiBit );
#ifdef WATCH_PIPE_2
  printf("ID= %4d Corder= %2d uiBit= %3d\n", CWID, m_uiPIPEId, uiBit);
  if(CWID == 436){
    CWID = CWID;
  }
  CWID++;
#endif //WATCH_PIPE_2
#ifdef PIPE_V2F_8BITS
  ui64CW_Top    = m_pauiCodeword[uiBit>>4];
  ui64CW_Bottom = m_pauiCodeword[uiBit&0xf];
  m_uiValue   = UInt( ui64CW_Top >> 6 );
  m_uiLength  = UInt( ui64CW_Bottom & 63 );
  m_uiValue   = (m_uiValue << m_uiLength) + UInt( ui64CW_Bottom >> 6 );
  m_uiLength  += UInt( ui64CW_Top & 63 );
#else //PIPE_V2F_8BITS
  ui64CW_Top    = m_pauiCodeword[uiBit];
  m_uiValue   = UInt( ui64CW_Top >> 6 );
  m_uiLength  = UInt( ui64CW_Top & 63 );
#endif //PIPE_V2F_8BITS
}

#else //PIPE_V2V_V2F

Void
V2VDecoder::xReadNextCodeword()
{
  UInt    uiBit   = 0;
  UInt    uiState = 0;
  UInt64  ui64CW  = 0;
  do
  {
    m_pcTComBitstream->read( 1, uiBit );
    uiState = m_paucStateTransition [ ( uiState << 2 ) + uiBit ];
    ui64CW  = m_pauiCodeword        [ uiState ];
  } while( ui64CW == 0 );
  m_uiValue   = UInt( ui64CW >> 6 );
  m_uiLength  = UInt( ui64CW & 63 );
}
#endif //PIPE_V2V_V2F


TDecBinPIPE::TDecBinPIPE()
: m_pacState2Idx( TDecPIPETables::sm_State2Idx )
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
#if PIPE_LOW_DELAY_OPTION
    m_apcV2VDecoders[ uiIdx ] = new V2VDecoder( uiIdx, &m_uiMaxDelayBitCounter );
#else
    m_apcV2VDecoders[ uiIdx ] = new V2VDecoder( uiIdx );
#endif
  }
}

TDecBinPIPE::~TDecBinPIPE()
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    delete m_apcV2VDecoders[ uiIdx ];
  }
}

Void
TDecBinPIPE::initDelay( UInt uiMaxDelayInBits )
{
#if PIPE_LOW_DELAY_OPTION
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_apcV2VDecoders[ uiIdx ]->initDelay( uiMaxDelayInBits );
  }
#endif
}


Void
TDecBinPIPE::init( TComBitstream* pcTComBitstream )
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_apcV2VDecoders[ uiIdx ]->setBitstream( pcTComBitstream );
  }
}


Void
TDecBinPIPE::uninit()
{
}


Void
TDecBinPIPE::start()
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_apcV2VDecoders[ uiIdx ]->start();
  }
#if PIPE_LOW_DELAY_OPTION
  m_uiMaxDelayBitCounter = 0;
#endif
}


Void
TDecBinPIPE::finish()
{
}

#ifdef WATCH_PIPE
UInt DecID = 0;
#endif //WATCH_PIPE


Void
TDecBinPIPE::decodeBin( UInt& ruiBin, ContextModel &rcCtxModel )
{
#ifdef GET_TIME_ENTROPY
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_ENTROPY
  UInt uiPIPEId = m_pacState2Idx  [ rcCtxModel.getState() ];
  UInt uiBin    = m_apcV2VDecoders[ uiPIPEId ]->getNextBin();
  ruiBin        = ( uiBin ^ rcCtxModel.getMps() );
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d\n", DecID, rcCtxModel.getState(), rcCtxModel.getMps(), ruiBin);
  if(DecID == 8240){
    DecID = DecID;
  }
  if(DecID == 58713){
    DecID = DecID;
  }
  DecID++;
#endif //WATCH_PIPE

#ifdef GET_TIME_BIT
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_BIT

  if(  uiBin )
  {
    rcCtxModel.updateLPS();
  }
  else
  {
    rcCtxModel.updateMPS();
  }

#ifdef GET_TIME_BIT
  getrusage(RUSAGE_SELF, &t2);
  DecBitTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecBitTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_BIT

#ifdef GET_TIME_ENTROPY
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_ENTROPY
}


Void
TDecBinPIPE::decodeBinEP( UInt& ruiBin )
{
#ifdef GET_TIME_ENTROPY
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_ENTROPY
#ifdef GET_TIME_EP
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_EP

  ruiBin = m_apcV2VDecoders[ 0 ]->getNextBin();
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d EP\n", DecID, 0, 0, ruiBin);
  if(DecID == 58713){
    DecID = DecID;
  }
  DecID++;
#endif //WATCH_PIPE

#ifdef GET_TIME_ENTROPY
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_ENTROPY
#ifdef GET_TIME_EP
  getrusage(RUSAGE_SELF, &t2);
  DecEPTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecEPTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_EP
}

#ifdef PIPE_V2F_BYPASS_MULTIREAD
Void
TDecBinPIPE::decodeBinEP_multi( UInt& ruiBin, UInt uiLength )
{
#ifdef GET_TIME_ENTROPY
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_ENTROPY
#ifdef GET_TIME_EP
  struct rusage t1, t2;
  getrusage(RUSAGE_SELF, &t1);
#endif //GET_TIME_EP

  ruiBin = m_apcV2VDecoders[ 0 ]->getNextBin_EP_multi( uiLength );
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d EP\n", DecID, 0, 0, ruiBin);
  if(DecID == 58713){
    DecID = DecID;
  }
  DecID++;
#endif //WATCH_PIPE
#ifdef GET_TIME_ENTROPY
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecEntTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_ENTROPY
#ifdef GET_TIME_EP
  getrusage(RUSAGE_SELF, &t2);
  DecEPTime += t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
  getrusage(RUSAGE_SELF, &t1);
  getrusage(RUSAGE_SELF, &t2);
  DecEPTime -= t2.ru_utime.tv_sec + (double)t2.ru_utime.tv_usec*1e-6 - t1.ru_utime.tv_sec - (double)t1.ru_utime.tv_usec*1e-6;
#endif //GET_TIME_EP
}
#endif //PIPE_V2F_BYPASS_MULTIREAD


Void
TDecBinPIPE::decodeBinTrm( UInt& ruiBin )
{
  ruiBin = m_apcV2VDecoders[ NUM_V2V_CODERS - 1 ]->getNextBin();
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d Tem\n", DecID, 0, 0, ruiBin);
  if(DecID == 58713){
    DecID = DecID;
  }
  DecID++;
#endif //WATCH_PIPE
}

