/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, FRAUNHOFER HHI
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of FRAUNHOFER HHI
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TEncBinCoderPIPE.cpp
    \brief    binary entropy encoder for PIPE
*/

#include "TEncBinCoderPIPE.h"


#define INVALID  UInt(-1)

V2VEncoder::V2VEncoder( const UInt uiPIPEId, CodewordBuffer& rcCodewordBuffer )
: m_uiPIPEId            ( uiPIPEId )
, m_paucStateTransition ( TEncPIPETables::sm_StateTransition[ m_uiPIPEId ] )
, m_pauiCodeword        ( TEncPIPETables::sm_Codeword       [ m_uiPIPEId ] )
, m_rcCodewordBuffer    ( rcCodewordBuffer )
#ifdef PIPE_V2F_8BITS
, m_cCWState            ( 0 )
#endif //PIPE_V2F_8BITS
{
}

Void
V2VEncoder::start()
{
  m_uiState = 0;
#ifdef PIPE_V2F_8BITS
  m_cCWState = 0;
#endif //PIPE_V2F_8BITS
}


#ifdef PIPE_V2F_8BITS
Void
V2VEncoder::pushBackV2F( UInt uiBin )
{
  if( m_uiPIPEId == PIPE_TRM_CODER ){
    if( m_uiState == 0 )
    {
      m_rcCodewordBuffer.reserve( m_uiPIPEId );
    }
    if( ! uiBin ) m_uiState++;
    if( uiBin || m_uiState == 255){
      m_rcCodewordBuffer.write( PIPE_TRM_CODER, m_uiState, 8 );
      m_uiState = 0;  
    }
  }
  else
  {
    if( m_cCWState == 0 )
    {
      m_cCWState = 1;
      m_rcCodewordBuffer.reserve( m_uiPIPEId );
    }
    m_uiState     = m_paucStateTransition[ ( m_uiState << 2 ) + uiBin ];
    UInt64 ui64CW = m_pauiCodeword[ m_uiState ];
    if( ui64CW )
    {
      m_rcCodewordBuffer.writeV2F( m_uiPIPEId, UInt( ui64CW >> 6 ));
      m_uiState = 0;
      m_cCWState = (m_cCWState << 1) & 0x03;
    }
  }
}
#endif //PIPE_V2F_8BITS


Void
V2VEncoder::pushBack( UInt uiBin )
{
  if( m_uiState == 0 )
  {
    m_rcCodewordBuffer.reserve( m_uiPIPEId );
  }
  m_uiState     = m_paucStateTransition[ ( m_uiState << 2 ) + uiBin ];
  UInt64 ui64CW = m_pauiCodeword[ m_uiState ];
  if( ui64CW )
  {
    m_rcCodewordBuffer.write( m_uiPIPEId, UInt( ui64CW >> 6 ), UInt( ui64CW & 63 ) );
    m_uiState = 0;
  }
}


#ifdef PIPE_V2F_8BITS
Void
V2VEncoder::flushV2F()
{
  if (m_uiPIPEId == PIPE_TRM_CODER){
    if(m_uiState) m_rcCodewordBuffer.writeV2F( PIPE_TRM_CODER, 255 );
  }
  else if( m_cCWState )
  {
    UInt64 ui64CW = m_pauiCodeword[ m_paucStateTransition [ ( m_uiState << 2 ) + 2 ] ];
    m_rcCodewordBuffer.writeV2F( m_uiPIPEId, UInt( ui64CW >> 6 ) );
    if( m_cCWState == 1 ){
      m_rcCodewordBuffer.writeV2F( m_uiPIPEId, 0 );
      m_cCWState = 0;
    }
  }
  m_uiState = 0;
}
#endif //PIPE_V2F_8BITS


Void
V2VEncoder::flush()
{
  if( m_uiState )
  {
    UInt64 ui64CW = m_pauiCodeword[ m_paucStateTransition [ ( m_uiState << 2 ) + 2 ] ];
    m_rcCodewordBuffer.write( m_uiPIPEId, UInt( ui64CW >> 6 ), UInt( ui64CW & 63 ) );
    m_uiState = 0;
  }
}


CodewordBuffer::CodewordBuffer()
: m_pcTComBitIf       ( 0 )
, m_uiBufferSize      ( INIT_CODEWORD_BUFFER_SIZE )
, m_pacCodewordBuffer ( new CBufEntry [ m_uiBufferSize ] )
#if PIPE_LOW_DELAY_OPTION
, m_papcV2VEncoders   ( 0 )
#endif
{
}

CodewordBuffer::~CodewordBuffer()
{
  delete [] m_pacCodewordBuffer;
}

#if PIPE_LOW_DELAY_OPTION
Void
CodewordBuffer::initDelay( V2VEncoder** papcV2VEncoder, UInt uiMaxDelayInBits )
{
  m_papcV2VEncoders             = papcV2VEncoder;
  m_uiMaxOverallCodewordLength  = 0;
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_auiMaxCodewordLength    [ uiIdx ] = TEncPIPETables::sm_MaxTableSize[ uiIdx ];
    if( m_auiMaxCodewordLength[ uiIdx ] > m_uiMaxOverallCodewordLength )
    {
      m_uiMaxOverallCodewordLength      = m_auiMaxCodewordLength[ uiIdx ];
    }
  }
  m_uiMaxDelayInBits            = ( uiMaxDelayInBits ? uiMaxDelayInBits : UInt(-1) );
  assert( m_uiMaxDelayInBits   >= m_uiMaxOverallCodewordLength );
}
#endif

Void
CodewordBuffer::start()
{
  m_uiNextWriteIdx                          = 0;
  m_uiNextReserveIdx                        = 0;
  m_uiNumFreeEntries                        = m_uiBufferSize;
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_auiReservedIdBuffer[ uiIdx ]          = INVALID;
#ifdef PIPE_V2F_8BITS
    m_cTopCWFlag[ uiIdx ] = 1;
#endif //PIPE_V2F_8BITS
  }
  for( UInt uiIdx = 0; uiIdx < m_uiBufferSize; uiIdx++ )
  {
    m_pacCodewordBuffer[ uiIdx ].ucCWLength = 0;
  }
#if PIPE_LOW_DELAY_OPTION
  m_uiMaxDelayBitCounter                    = 0;
#endif
}

Void
CodewordBuffer::finish()
{
  m_pcTComBitIf->write( 1, 1 );
}

Void
CodewordBuffer::reserve( UInt uiPIPEId )
{
#if PIPE_LOW_DELAY_OPTION // low-delay encoder flush
#ifdef PIPE_V2F_8BITS
  UInt uiMaxBitCounter = m_uiMaxDelayInBits - 8;
#else //PIPE_V2F_8BITS
  UInt uiMaxBitCounter = m_uiMaxDelayInBits - m_auiMaxCodewordLength[ uiPIPEId ];
#endif //PIPE_V2F_8BITS
  while( m_uiMaxDelayBitCounter > uiMaxBitCounter )
  {
#ifdef PIPE_V2F_8BITS
    m_papcV2VEncoders[ m_pacCodewordBuffer[ m_uiNextWriteIdx ].ucPIPEId ]->flushV2F();
#else //PIPE_V2F_8BITS
    m_papcV2VEncoders[ m_pacCodewordBuffer[ m_uiNextWriteIdx ].ucPIPEId ]->flush();
#endif //PIPE_V2F_8BITS
  }
#endif
  if( m_uiNumFreeEntries == 0 )
  {
    xIncreaseBufferSize();
  }
  m_pacCodewordBuffer   [ m_uiNextReserveIdx ].ucPIPEId = uiPIPEId;
  m_auiReservedIdBuffer [ uiPIPEId           ]          = m_uiNextReserveIdx++;
  if( m_uiNextReserveIdx == m_uiBufferSize )
  {
    m_uiNextReserveIdx = 0;
  }
  m_uiNumFreeEntries--;
#if PIPE_LOW_DELAY_OPTION
#ifdef PIPE_V2F_8BITS
  m_uiMaxDelayBitCounter += 8;
#else //PIPE_V2F_8BITS
  m_uiMaxDelayBitCounter += m_auiMaxCodewordLength[ uiPIPEId ];
#endif //PIPE_V2F_8BITS
#endif
}


#ifdef WATCH_PIPE_2
UInt CWID = 0;
#endif //WATCH_PIPE_2


#ifdef PIPE_V2F_8BITS
Void
CodewordBuffer::writeV2F( UInt uiPIPEId, UInt uiValue)
{
  if (uiPIPEId == PIPE_TRM_CODER) {
    UInt uiCurrIdx = m_auiReservedIdBuffer[ uiPIPEId ];
    m_auiReservedIdBuffer[ PIPE_TRM_CODER ]               = INVALID;
    m_pacCodewordBuffer  [ uiCurrIdx ].uiCWValue          = uiValue;
    m_pacCodewordBuffer  [ uiCurrIdx ].ucCWLength         = 8;
  }
  else
  {
    if(m_cTopCWFlag[ uiPIPEId           ] == 1){
      m_cTopCWFlag[ uiPIPEId           ] = 0;
      UInt uiCurrIdx = m_auiReservedIdBuffer[ uiPIPEId ];
      m_pacCodewordBuffer  [ uiCurrIdx ].uiCWValue  = uiValue;
      //m_pacCodewordBuffer  [ uiCurrIdx ].ucCWLength   = 4;
    }
    else{
      m_cTopCWFlag[ uiPIPEId           ] = 1;
      UInt uiCurrIdx = m_auiReservedIdBuffer[ uiPIPEId ];
      m_auiReservedIdBuffer[ uiPIPEId  ]                  = INVALID;
      m_pacCodewordBuffer  [ uiCurrIdx ].uiCWValue  = (m_pacCodewordBuffer  [ uiCurrIdx ].uiCWValue << 4) + uiValue;
      m_pacCodewordBuffer  [ uiCurrIdx ].ucCWLength         = 8;
      if( uiCurrIdx == m_uiNextWriteIdx )
      {
        xWriteV2F();
      }    
    }
  }
}
#endif //PIPE_V2F_8BITS

Void
CodewordBuffer::write( UInt uiPIPEId, UInt uiValue, UInt uiLength )
{
  UInt uiCurrIdx = m_auiReservedIdBuffer[ uiPIPEId ];
  m_auiReservedIdBuffer[ uiPIPEId  ]            = INVALID;
  m_pacCodewordBuffer  [ uiCurrIdx ].ucCWLength = uiLength;
  m_pacCodewordBuffer  [ uiCurrIdx ].uiCWValue  = uiValue;
  if( uiCurrIdx == m_uiNextWriteIdx )
  {
    xWrite();
  }
}

Void
CodewordBuffer::xIncreaseBufferSize()
{
  UInt              uiNewFreeEntries      = m_uiNumFreeEntries  + m_uiBufferSize;
  UInt              uiNewBufferSize       = m_uiBufferSize      + m_uiBufferSize;
  UInt              uiNewNextWriteIdx     = 0;
  UInt              uiNewNextReservedIdx  = m_uiBufferSize      - m_uiNumFreeEntries;
  CBufEntry*        pacNewCodewordBuffer  = new CBufEntry       [ uiNewBufferSize ];
  for( UInt uiSrcIdx = m_uiNextWriteIdx, uiDesIdx = uiNewNextWriteIdx; uiDesIdx < uiNewNextReservedIdx; uiDesIdx++, uiSrcIdx = ( uiSrcIdx + 1 ) % m_uiBufferSize )
  {
    pacNewCodewordBuffer[ uiDesIdx ] = m_pacCodewordBuffer[ uiSrcIdx ];
    if( pacNewCodewordBuffer[ uiDesIdx ].ucCWLength == 0 )
    {
      m_auiReservedIdBuffer[ pacNewCodewordBuffer[ uiDesIdx ].ucPIPEId ] = uiDesIdx;
    }
  }
  for( UInt uiDesIdx = uiNewNextReservedIdx; uiDesIdx < uiNewBufferSize; uiDesIdx++ )
  {
    pacNewCodewordBuffer[ uiDesIdx ].ucCWLength = 0;
  }
  delete [] m_pacCodewordBuffer;
  m_uiNextWriteIdx    = uiNewNextWriteIdx;
  m_uiNextReserveIdx  = uiNewNextReservedIdx;
  m_uiBufferSize      = uiNewBufferSize;
  m_uiNumFreeEntries  = uiNewFreeEntries;
  m_pacCodewordBuffer = pacNewCodewordBuffer;
}

Void
CodewordBuffer::xWrite()
{
  while( m_pacCodewordBuffer[ m_uiNextWriteIdx ].ucCWLength )
  {
    CBufEntry&  rcEntry     = m_pacCodewordBuffer[ m_uiNextWriteIdx ];
    m_pcTComBitIf->write( rcEntry.uiCWValue, rcEntry.ucCWLength );
#if PIPE_LOW_DELAY_OPTION
    m_uiMaxDelayBitCounter -= m_auiMaxCodewordLength[ rcEntry.ucPIPEId ];
#endif
    rcEntry.ucCWLength      = 0;
    m_uiNextWriteIdx++;
    if( m_uiNextWriteIdx   == m_uiBufferSize )
    {
      m_uiNextWriteIdx      = 0;
    }
    m_uiNumFreeEntries++;
  }
}


#ifdef PIPE_V2F_8BITS
Void
CodewordBuffer::xWriteV2F()
{
  while( m_pacCodewordBuffer[ m_uiNextWriteIdx ].ucCWLength )
  {
    CBufEntry&  rcEntry     = m_pacCodewordBuffer[ m_uiNextWriteIdx ];
    m_pcTComBitIf->write( rcEntry.uiCWValue, 8 );
#ifdef WATCH_PIPE_2
    printf("ID= %4d Corder= %2d uiBit= %3d\n", CWID, rcEntry.ucPIPEId, rcEntry.uiCWValue);
    if(CWID == 436){
      CWID = CWID;
    }
    CWID++;
#endif //WATCH_PIPE_2

#if PIPE_LOW_DELAY_OPTION
    m_uiMaxDelayBitCounter -= 8;
#endif
    rcEntry.ucCWLength      = 0;
    m_uiNextWriteIdx++;
    if( m_uiNextWriteIdx   == m_uiBufferSize )
    {
      m_uiNextWriteIdx      = 0;
    }
    m_uiNumFreeEntries++;
  }
}
#endif //PIPE_V2F_8BITS


TEncBinPIPE::TEncBinPIPE()
: m_pacStat2Idx     ( TEncPIPETables::sm_State2Idx )
, m_cCodewordBuffer ()
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_apcV2VEncoders[ uiIdx ] = new V2VEncoder( uiIdx, m_cCodewordBuffer );
  }
#if PIPE_LOW_DELAY_OPTION
  m_cCodewordBuffer.initDelay( m_apcV2VEncoders, 0 );
#endif
}

TEncBinPIPE::~TEncBinPIPE()
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    delete m_apcV2VEncoders[ uiIdx ];
  }
}

Void
TEncBinPIPE::initDelay( UInt uiMaxDelayInBits )
{
#if PIPE_LOW_DELAY_OPTION
  m_cCodewordBuffer.initDelay( m_apcV2VEncoders, uiMaxDelayInBits );
#endif
}

Void
TEncBinPIPE::init( TComBitIf* pcTComBitIf )
{
  m_cCodewordBuffer.setBitStream( pcTComBitIf );
}

Void
TEncBinPIPE::uninit()
{
}

Void
TEncBinPIPE::start()
{
  m_cCodewordBuffer.start();
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
    m_apcV2VEncoders[ uiIdx ]->start();
  }
}

Void
TEncBinPIPE::finish()
{
  for( UInt uiIdx = 0; uiIdx < NUM_V2V_CODERS; uiIdx++ )
  {
#ifdef PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiIdx ]->flushV2F();
#else //PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiIdx ]->flush();
#endif //PIPE_V2F_8BITS
  }
  m_cCodewordBuffer.finish();
}

Void
TEncBinPIPE::copyState( TEncBinIf* pcTEncBinIf )
{
  assert(0);
}

Void  
TEncBinPIPE::resetBits()
{
  assert(0); 
}

UInt
TEncBinPIPE::getNumWrittenBits()
{
  assert(0);
  return 0;
}

#ifdef WATCH_PIPE
UInt ID = 0;
#endif //WATCH_PIPE

Void
TEncBinPIPE::encodeBin( UInt uiBin, ContextModel &rcCtxModel )
{
#if HHI_RQT
  {
    DTRACE_CABAC_V( g_nSymbolCounter++ )
    DTRACE_CABAC_T( "\tstate=" )
    DTRACE_CABAC_V( ( rcCtxModel.getState() << 1 ) + rcCtxModel.getMps() )
    DTRACE_CABAC_T( "\tsymbol=" )
    DTRACE_CABAC_V( uiBin )
    DTRACE_CABAC_T( "\n" )
  }
#endif

#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d\n", ID, rcCtxModel.getState(), rcCtxModel.getMps(), uiBin);
  if(ID == 8231){
    ID = ID;
  }
  ID++;
#endif //WATCH_PIPE

  UInt uiPIPEId = m_pacStat2Idx[ rcCtxModel.getState() ];
  if( uiBin != rcCtxModel.getMps() )
  {
#ifdef PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiPIPEId ]->pushBackV2F( 1 );
#else //PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiPIPEId ]->pushBack( 1 );
#endif //PIPE_V2F_8BITS
    rcCtxModel.updateLPS();
  }
  else
  {
#ifdef PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiPIPEId ]->pushBackV2F( 0 );
#else //PIPE_V2F_8BITS
    m_apcV2VEncoders[ uiPIPEId ]->pushBack( 0 );
#endif //PIPE_V2F_8BITS
    rcCtxModel.updateMPS();
  }
}

Void
TEncBinPIPE::encodeBinEP( UInt uiBin )
{
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d EP\n", ID, 0, 0, uiBin);
  if(ID == 8231){
    ID = ID;
  }
  ID++;
#endif //WATCH_PIPE

#ifdef PIPE_V2F_8BITS
  m_apcV2VEncoders[ 0 ]->pushBackV2F( uiBin );
#else //PIPE_V2F_8BITS
  m_apcV2VEncoders[ 0 ]->pushBack( uiBin );
#endif //PIPE_V2F_8BITS
}


Void
TEncBinPIPE::encodeBinTrm( UInt uiBin )
{
#ifdef WATCH_PIPE
  printf("ID= %6d, State= %2d, MPS= %d, uiBin= %d Trm\n", ID, 0, 0, uiBin);
  if(ID == 8231){
    ID = ID;
  }
  ID++;
#endif //WATCH_PIPE

#ifdef PIPE_V2F_8BITS
  m_apcV2VEncoders[ NUM_V2V_CODERS - 1 ]->pushBackV2F( uiBin );
#else //PIPE_V2F_8BITS
  m_apcV2VEncoders[ NUM_V2V_CODERS - 1 ]->pushBack( uiBin );
#endif //PIPE_V2F_8BITS
}

